#include <iostream>
#include <mpir.h>

#include "Constants.h"
#include "Utility.h"
#include "Timer.h"
#include "C_API.h"

using namespace std;

void compute_and_print_pi_c()
{
    cout << "C API: pi to " << commafy(PLACES) << " places:" << endl;

    // Time the computation of pi.
    mpf_t pi;
    long ns = Timer::time_it(compute_pi_c, pi);

    // Convert the number to a string for printing.
    mp_exp_t exp;      // exponent (not used)
    char *str = NULL;
    char *s = mpf_get_str(str, &exp, BASE, PRECISION, pi);

    print(s, ns);
}

void compute_pi_c(mpf_t& pi)
{
    mpf_set_default_prec(BIT_COUNT*PRECISION/2);  // precision in bits

    mpf_t one, two, four, six, sqrt2, sqrt2x4;
    mpf_t y, y_prev, a, a_prev, y4, yRoot4, aTerm, powers2, y2;

    // Multiple-precision constants.
    mpf_init(one);  mpf_set_str(one,  "1", BASE);
    mpf_init(two);  mpf_set_str(two,  "2", BASE);
    mpf_init(four); mpf_set_str(four, "4", BASE);
    mpf_init(six);  mpf_set_str(six,  "6", BASE);

    // The square root of 2 and four times the square root of 2.
    mpf_init(sqrt2);   mpf_sqrt(sqrt2, two);
    mpf_init(sqrt2x4); mpf_mul(sqrt2x4, four, sqrt2);

    // Multiple-precision variables.
    mpf_init(y);
    mpf_init(y_prev); mpf_sub(y_prev, sqrt2, one);
    mpf_init(a);
    mpf_init(a_prev); mpf_sub(a_prev, six, sqrt2x4);
    mpf_init(y4);
    mpf_init(yRoot4);
    mpf_init(aTerm);
    mpf_init(powers2); mpf_set_str(powers2, "2", BASE);
    mpf_init(y2);

    // Loop MAX_ITERATIONS times.
    for (int i = 1; i <= MAX_ITERATIONS; i++)
    {
        // Compute y^4
        power4(y4, y_prev);

        mpf_t temp1, temp2;
        mpf_init(temp1);
        mpf_init(temp2);

        // Compute the fourth root of (1 - y^4)
        mpf_sub(temp1, one, y4);
        root4(yRoot4, temp1);

        // Compute y
        mpf_sub(temp1, one, yRoot4);
        mpf_add(temp2, one, yRoot4);
        mpf_div(y, temp1, temp2);

        // Compute a*(1 + y)^4
        mpf_add(temp1, one, y);
        power4(temp2, temp1);
        mpf_mul(aTerm, a_prev, temp2);

        // Compute 2^(2*i + 1)
        mpf_mul(powers2, four, powers2);

        // Compute y^2
        mpf_mul(y2, y, y);

        // Compute a
        mpf_add(temp1, one, y);
        mpf_add(temp1, temp1, y2);
        mpf_mul(temp1, temp1, y);
        mpf_mul(temp1, temp1, powers2);
        mpf_sub(a, aTerm, temp1);

        mpf_set(a_prev, a);
        mpf_set(y_prev, y);
    }

    // Compute pi = 1/a
    mpf_init(pi);
    mpf_div(pi, one, a);
}

void root4(mpf_t& root4, const mpf_t& x)
{
    mpf_t root2;
    mpf_init(root2);

    mpf_sqrt(root2, x);
    mpf_sqrt(root4, root2);
}

void power4(mpf_t& xxxx, const mpf_t& x)
{
    mpf_t xx;
    mpf_init(xx);

    mpf_mul(xx, x, x);
    mpf_mul(xxxx, xx, xx);
}
